"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def test_stft():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        model  # unused # noqa
        x = extern_data["data"]
        out, out_spatial_dim, out_dim = rf.stft(x, in_spatial_dim=time_dim, frame_step=3, frame_length=5, fft_length=6)
        out.mark_as_default_output(shape=(batch_dim, out_spatial_dim, out_dim))
        out, out_spatial_dim, out_dim = rf.stft(x, in_spatial_dim=time_dim, frame_step=3, frame_length=32)
        out.mark_as_output("zero_len", shape=(batch_dim, out_spatial_dim, out_dim))
        out, out_spatial_dim, out_dim = rf.stft(x, in_spatial_dim=time_dim, frame_step=3, frame_length=5)
        out.mark_as_output("normal", shape=(batch_dim, out_spatial_dim, out_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_stft_after_padding():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    pad_left_dim = Dim(Tensor("pad_left", [batch_dim], dtype="int32"))
    pad_right_dim = Dim(Tensor("pad_right", [batch_dim], dtype="int32"))
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="float32"),
            "left": Tensor("left", [batch_dim, pad_left_dim], dtype="float32"),
            "right": Tensor("right", [batch_dim, pad_right_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_):
        x, left, right = extern_data["data"], extern_data["left"], extern_data["right"]
        x, spatial_dim = rf.concat((left, pad_left_dim), (x, time_dim), (right, pad_right_dim))
        assert x.raw_tensor.shape[x.get_axis_from_description(spatial_dim)] == spatial_dim.get_dim_value()
        out, out_spatial_dim, out_dim = rf.stft(
            x, in_spatial_dim=spatial_dim, frame_step=3, frame_length=17, fft_length=32
        )
        assert out.raw_tensor.shape[out.get_axis_from_description(out_spatial_dim)] == out_spatial_dim.get_dim_value()
        out.mark_as_default_output(shape=(batch_dim, out_spatial_dim, out_dim))

    run_model(
        extern_data,
        lambda **_: rf.Module(),
        _forward_step,
        dyn_dim_min_sizes={time_dim: 50},
        dyn_dim_max_sizes={time_dim: 100, pad_left_dim: 4, pad_right_dim: 4},
        test_tensorflow=False,
        test_single_batch_entry=False,
    )


def test_mel_filterbank():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    feat_dim = Dim(10, name="mel")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, source: rf.Tensor, *, in_spatial_dim: Dim) -> Tuple[Tensor, Dim]:
            # log mel filterbank features
            source, in_spatial_dim, in_dim_ = rf.stft(
                source, in_spatial_dim=in_spatial_dim, frame_step=3, frame_length=5, fft_length=6
            )
            source = rf.abs(source) ** 2.0
            source = rf.audio.mel_filterbank(source, in_dim=in_dim_, out_dim=feat_dim, sampling_rate=16)
            return source, in_spatial_dim

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, out_spatial_dim = model(extern_data["data"], in_spatial_dim=time_dim)
        out.mark_as_default_output(shape=(batch_dim, out_spatial_dim, feat_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)
    # Run again to specifically test the caching logic.
    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_audio_specaugment():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32", feature_dim=in_dim),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        model  # noqa  # unused
        data = extern_data["data"]
        out = rf.audio.specaugment(data, spatial_dim=time_dim, feature_dim=in_dim, only_on_train=False)
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations

import _setup_test_env  # noqa

import sys
import unittest

from returnn.util import better_exchook
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def test_neg():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return -x

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_compare_bc():
    beam_dim = Dim(3, name="beam")
    in_dim = Dim(7, name="in")
    extern_data = TensorDict({"idx": Tensor("idx", [batch_dim, beam_dim], dtype="int32", sparse_dim=in_dim)})

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        idx = extern_data["idx"]
        cond = rf.compare_bc(idx, "!=", rf.range_over_dim(in_dim))
        cond.mark_as_default_output(shape=(batch_dim, beam_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_logical_or():
    beam_dim = Dim(3, name="beam")
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "a": Tensor("a", [batch_dim, beam_dim], dtype="bool"),
            "b": Tensor("b", [batch_dim, beam_dim, in_dim], dtype="bool"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        a, b = extern_data["a"], extern_data["b"]
        cond = a | b
        cond.mark_as_default_output(shape=(batch_dim, beam_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_squared_difference():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "a": Tensor("a", [batch_dim, time_dim, in_dim], dtype="float32"),
            "b": Tensor("b", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, a: Tensor, b: Tensor) -> Tensor:
            return rf.squared_difference(a, b)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["a"], extern_data["b"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_log_add_exp():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "a": Tensor("a", [batch_dim, time_dim, in_dim], dtype="float32"),
            "b": Tensor("b", [in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_):
        out = rf.log_add_exp(extern_data["a"], extern_data["b"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda **_: rf.Module(), _forward_step)


def test_abs_complex():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="complex64"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.abs(x)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_relu():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.relu(x)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_log_sigmoid():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        out = rf.log_sigmoid(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_cumsum():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="int32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        out = rf.cumsum(extern_data["data"], spatial_dim=time_dim)
        out.mark_as_default_output(shape=(batch_dim, time_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



from __future__ import annotations

import tensorflow as tf
import returnn.tf.compat as tf_compat
from returnn.tf.layers.basic import LayerBase, _ConcatInputLayer, get_concat_sources_data_template


# ---------- Utilities ----------


def batch_sizes_after_windowing(sizes, window):
    """
    :param tf.Tensor sizes: (batch_sizes)
    :param int window: size of the applied window
    :return: sizes for each batch after applying a window on each batch
    :rtype: tf.Tensor
    """

    def fold_times(acc, x):
        r1 = tf.tile([window], [tf.maximum(x - window + 1, 0)])
        r2 = tf.range(tf.minimum(x, window - 1), 0, -1)
        return tf.concat([acc, r1, r2], 0)

    return tf.foldl(
        fold_times,
        sizes,
        tf_compat.v1.placeholder_with_default(tf.zeros([0], dtype="int32"), [None]),
        name="fold_sizes",
    )


def batch_indices_after_windowing(sizes, window):
    """
    here we compute the start and end times for each of the new batches when applying a window
    :param tf.Tensor sizes: (batch_sizes)
    :param int window: size of the applied window
    :return: tensor of shape (?, 3), contains batch index, start-frame and end-frame for each batch after applying a window
    :rtype: tf.Tensor
    """

    def fold_batches(acc, x):
        b = x[0]
        l = x[1]
        batch = tf.tile([b], [l])
        start = tf.range(l)
        end = tf.minimum(tf.range(window, l + window), l)
        return tf.concat([acc, tf.transpose(tf.stack([batch, start, end]))], axis=0)

    return tf.foldl(
        fold_batches,
        tf.stack([tf.range(tf.shape(sizes)[0]), sizes], axis=1),
        tf_compat.v1.placeholder_with_default(tf.zeros([0, 3], dtype="int32"), [None, 3]),
        name="fold_batches",
    )


# ---------- Layers ----------


class SegmentInputLayer(_ConcatInputLayer):
    """
    This layer takes the input data, applies a window and outputs each window as a new batch, this is more
    efficient than a window as a new dimension if sequences have varying lengths
    """

    layer_class = "segment_input"

    def __init__(self, window=15, **kwargs):
        super(SegmentInputLayer, self).__init__(**kwargs)
        sizes = self.input_data.size_placeholder[0]
        new_sizes = batch_sizes_after_windowing(sizes, window)

        def fold_data(acc, x):
            batch_idx = x[0]
            num_frames = x[1]
            res = tf.expand_dims(tf.range(num_frames), -1)  # start times
            res = tf.tile(res, [1, window])  # fill add time dimension
            res += tf.range(window)  # add offsets
            res = tf.where(res >= num_frames, tf.zeros_like(res), res)  # filter frames that go past the end
            if self.input_data.is_batch_major:
                res = tf.stack([tf.ones_like(res) * batch_idx, res], axis=2)  # add batch_index in first dim
            else:
                res = tf.stack([res, tf.ones_like(res) * batch_idx], axis=2)  # add batch_index in second dim
            return tf.concat([acc, res], 0)

        initial = tf_compat.v1.placeholder_with_default((tf.zeros([0, window, 2], dtype=tf.int32)), [None, window, 2])
        indices = tf.foldl(
            fold_data, tf.stack([tf.range(tf.shape(sizes)[0]), sizes], axis=1), initial, name="fold_data"
        )

        if self.input_data.is_time_major:
            indices = tf.transpose(indices, [1, 0, 2])

        self.output.placeholder = tf.gather_nd(self.input_data.placeholder, indices)
        self.output.size_placeholder[0] = new_sizes

    @classmethod
    def get_out_data_from_opts(cls, name, sources, window, **kwargs):
        out = get_concat_sources_data_template(sources, name="%s_output" % name)
        out.size_placeholder = {}
        out.size_placeholder[0] = None
        return out


class ClassesToSegmentsLayer(_ConcatInputLayer):
    """
    This layer takes a sequence of classes (=> sparse input) and applies a window (same as SegmentInput) to it.
    For each position t in the window it computes the relative frequencies of the classes up to and including
    that position t.
    """

    layer_class = "classes_to_segments"

    def __init__(self, num_classes, window=15, **kwargs):
        super(ClassesToSegmentsLayer, self).__init__(**kwargs)
        assert self.input_data.sparse

        sizes = self.input_data.size_placeholder[0]
        new_sizes = batch_sizes_after_windowing(sizes, window)
        batches = batch_indices_after_windowing(sizes, window)

        onehot = tf.one_hot(self.input_data.get_placeholder_as_batch_major(), num_classes)

        def compute(x):
            batch = x[0]
            start = x[1]
            end = x[2]

            padded_onehot = tf.concat(
                [onehot[batch][start:end], tf.zeros([tf.maximum(window - (end - start), 0), num_classes])], axis=0
            )
            classes = tf.cumsum(padded_onehot)
            normalization = tf.cast(tf.expand_dims(tf.range(1, window + 1), -1), classes.dtype)
            return classes / normalization

        self.output.placeholder = tf.map_fn(compute, batches, dtype="float32")
        self.output.size_placeholder[0] = new_sizes

    @classmethod
    def get_out_data_from_opts(cls, name, sources, num_classes, window, **kwargs):
        out = get_concat_sources_data_template(sources, name="%s_output" % name).copy_as_batch_major()
        out.size_placeholder = {}
        out.size_placeholder[0] = None
        out.sparse = False
        out.shape += (num_classes,)
        out.dtype = "float32"
        out.dim = num_classes
        return out


class ClassesToLengthDistributionLayer(_ConcatInputLayer):
    layer_class = "classes_to_length_distribution"

    def __init__(self, window=15, scale=1.0, **kwargs):
        super(ClassesToLengthDistributionLayer, self).__init__(**kwargs)
        assert self.input_data.sparse

        sizes = self.input_data.size_placeholder[0]
        batches = batch_indices_after_windowing(sizes, window)
        new_sizes = tf.fill((tf.shape(batches)[0],), 1)
        classes = self.input_data.get_placeholder_as_batch_major()

        def compute(bse):
            batch = bse[0]
            start = bse[1]
            end = bse[2]

            batch_cls = classes[batch][start:end]
            cls_not_eq = tf.not_equal(batch_cls[:-1], batch_cls[1:])
            cls_changed = tf.concat([cls_not_eq, [True]], axis=0)
            idx = tf.where(cls_changed)
            count = tf.squeeze(tf.concat([[idx[0] + 1], idx[1:] - idx[:-1]], axis=0), axis=1)
            freq = tf.cast(count, dtype="float32")

            res = tf.scatter_nd(idx, tf.cast(count, dtype="float32") / tf.cast(end - start, dtype="float32"), (window,))

            return res

        self.output.placeholder = tf.expand_dims(tf.map_fn(compute, batches, back_prop=False, dtype="float32"), axis=-2)
        self.output.size_placeholder[0] = new_sizes

    @classmethod
    def get_out_data_from_opts(cls, name, sources, window, **kwargs):
        out = get_concat_sources_data_template(sources, name="%s_output" % name).copy_as_batch_major()
        out.size_placeholder = {}
        out.size_placeholder[0] = None
        out.shape = (1, window)
        out.dim = window
        out.sparse = False
        out.dtype = "float32"
        return out


class ClassesToLengthDistributionGlobalLayer(_ConcatInputLayer):
    layer_class = "classes_to_length_distribution_global"

    def __init__(
        self, window=15, weight_falloff=1.0, target_smoothing=None, min_length=1, broadcast_axis="time", **kwargs
    ):
        super(ClassesToLengthDistributionGlobalLayer, self).__init__(**kwargs)
        assert self.input_data.sparse
        assert broadcast_axis in ["time", "feature"]

        sizes = self.input_data.size_placeholder[0]
        batches = batch_indices_after_windowing(sizes, window)
        new_sizes = tf.fill((tf.shape(batches)[0],), 1)
        classes = self.input_data.get_placeholder_as_batch_major()
        cls_not_eq = tf.not_equal(classes[:, :-1], classes[:, 1:])
        cls_changed = tf.concat([cls_not_eq, tf.fill((tf.shape(classes)[0], 1), True)], axis=1)

        lengths = tf.range(0.0, float(window), 1.0)
        end_distribution = tf.pow((1.0 - weight_falloff), lengths) * weight_falloff

        # add small weight at the last frame in case there is no ending label, then we want the last frame to be a label end
        # because the windows have different lengths the last frame might not be at index (window -1), but earlier, thus we
        # also add some zeros at the end
        no_label_backup = tf.convert_to_tensor([0.0] * (window - 1) + [1e-4] + [0.0] * (window - 1))

        # As the label ends are not well-defined in most cases, we allow the model a little bit of wiggle-room in the
        # decision where to put the label ends. This is achieved by smoothing the targets for one segment with a kernel
        # that is specified by target_smoothing.
        smoothing = None
        if target_smoothing is not None:
            import numpy as np

            assert len(target_smoothing) % 2 == 1

            smoothing = tf.TensorArray(
                dtype="float32", size=window, clear_after_read=False, infer_shape=False, name="smoothing_matrices"
            )

            s = sum(target_smoothing)
            target_smoothing = [v / s for v in target_smoothing]
            center = len(target_smoothing) // 2

            mat = np.zeros((window, window), dtype="float32")
            for i, v in enumerate(target_smoothing):
                mat += np.diag([v] * (window - abs(i - center)), i - center)

            for i in range(window):
                submat = np.diag([1.0] * window).astype("float32")
                submat[: i + 1, : i + 1] = mat[: i + 1, : i + 1]
                submat /= submat.sum(axis=1).reshape((-1, 1))
                smoothing = smoothing.write(i, submat)

        min_length_filter = tf.concat(
            [tf.zeros((min_length - 1,), dtype="float32"), tf.ones((window - min_length + 1), dtype="float32")], axis=0
        )

        def compute(bse):
            batch = bse[0]
            start = bse[1]
            end = bse[2]
            size = end - start

            cls_chg = cls_changed[batch][start:end]
            idx = tf.where(cls_chg)
            res = tf.scatter_nd(idx, end_distribution[: tf.shape(idx)[0]], (window,))
            if min_length > 1:
                res *= min_length_filter
            res += no_label_backup[window - size : 2 * window - size]
            res = res / tf.reduce_sum(res)
            if smoothing is not None:
                res = tf.tensordot(res, smoothing.read(size - 1), [[0], [0]])

            return res

        targets = tf.map_fn(compute, batches, back_prop=False, dtype="float32")
        if broadcast_axis == "time":
            self.output.placeholder = tf.expand_dims(targets, axis=-2)
        elif broadcast_axis == "feature":
            self.output.placeholder = tf.expand_dims(targets, axis=-1)
        self.output.size_placeholder[0] = new_sizes

    @classmethod
    def get_out_data_from_opts(cls, name, sources, window, broadcast_axis="time", **kwargs):
        out = get_concat_sources_data_template(sources, name="%s_output" % name).copy_as_batch_major()
        out.size_placeholder = {}
        out.size_placeholder[0] = None
        out.shape = (1, window) if broadcast_axis == "time" else (window, 1)
        out.dim = window if broadcast_axis == "time" else 1
        out.sparse = False
        out.dtype = "float32"
        return out


class SegmentAlignmentLayer(_ConcatInputLayer):
    layer_class = "segment_alignment"

    def __init__(self, num_classes, window=15, **kwargs):
        super(SegmentAlignmentLayer, self).__init__(**kwargs)
        assert self.input_data.sparse

        sizes = tf_compat.v1.div(self.input_data.size_placeholder[0], 2)
        batches = batch_indices_after_windowing(sizes, window)
        new_sizes = tf.fill((tf.shape(batches)[0],), window)

        input = self.input_data.get_placeholder_as_batch_major()
        input = tf.reshape(input, (tf.shape(input)[0], -1, 2))  # reinterpret last dimension as (dim, 2)
        end_distribution = tf.convert_to_tensor([1.0] + [0.0] * (window - 1))
        onehot = tf.one_hot(input[:, :, 0], num_classes)
        onehot = tf.pad(onehot, tf.constant([[0, 0], [0, window - 1], [0, 0]]), "CONSTANT")

        # add small weight at the last frame in case there is no ending label, then we want the last frame to be a label end
        # because the windows have different lengths the last frame might not be at index (window -1), but earlier, thus we
        # also add some zeros at the end
        no_label_backup = tf.constant([0.0] * (window - 1) + [1e-4] + [0.0] * (window - 1))

        def compute(bse):
            batch = bse[0]
            start = bse[1]
            end = bse[2]
            size = end - start

            seg_ended = input[batch, start:end, 1]
            idx = tf.where(tf.not_equal(seg_ended, 0))
            length_dist = tf.scatter_nd(idx, end_distribution[: tf.shape(idx)[0]], (window,))
            length_dist += no_label_backup[window - size : 2 * window - size]
            length_dist = length_dist / tf.reduce_sum(length_dist)
            length_dist = tf.expand_dims(length_dist, -1)

            result = onehot[batch, start : start + window, :] * length_dist
            return result

        targets = tf.map_fn(compute, batches, back_prop=False, dtype="float32")
        self.output.placeholder = targets
        self.output.size_placeholder[0] = new_sizes

    @classmethod
    def get_out_data_from_opts(cls, name, sources, num_classes, window, **kwargs):
        out = get_concat_sources_data_template(sources, name="%s_output" % name).copy_as_batch_major()
        out.size_placeholder = {}
        out.size_placeholder[0] = None
        out.shape = (window, num_classes)
        out.dim = num_classes
        out.sparse = False
        out.dtype = "float32"
        return out


class UnsegmentInputLayer(_ConcatInputLayer):
    """
    Takes the output of SegmentInput (sequences windowed over time and folded into batch-dim)
    and restores the original batch dimension. The feature dimension contains window * original_features
    many entries. The entries at time t all correspond to windows ending at time t. The window
    that started in the same frame comes first, then the window that started in the frame before and so on.
    This is also the format used for the segmental decoder in RASR.
    """

    layer_class = "unsegment_input"

    def __init__(self, **kwargs):
        super(UnsegmentInputLayer, self).__init__(**kwargs)
        sizes = self.input_data.size_placeholder[0]
        end_times = tf.squeeze(
            tf.where(tf.equal(sizes, 1)) + 1, axis=1
        )  # a batch of size one indicates that a sequence ended there
        start_times = tf.concat([[0], end_times[:-1]], axis=0)
        new_sizes = end_times - start_times
        max_size = tf.reduce_max(new_sizes)

        # first we shift the data in the time dimension (to get all windows that end at the same time into one batch)
        data = self.input_data.get_placeholder_as_time_major()

        def map_data(x):
            time = x[0]
            batches = x[1]
            size = tf.shape(batches)[0]
            out = tf.concat([batches[size - time :, :], batches[: size - time, :]], axis=0)
            return out

        data = tf.map_fn(map_data, [tf.range(tf.shape(data)[0]), data], dtype=self.input_data.dtype)

        # now we take the start and end times that we extracted above and feed them into extract_batch, which will get
        # the data for one of the original batches
        def extract_batch(x):
            start_batch = x[0]
            end_batch = x[1]
            # the next three lines are a convoluted way of writing data[:,start:end,:], but this notation did not work
            start = [tf.constant(0, dtype="int64"), start_batch, tf.constant(0, dtype="int64")]
            end = [tf.constant(-1, dtype="int64"), end_batch - start_batch, tf.constant(-1, dtype="int64")]
            d = tf.slice(data, start, end)
            d = tf.transpose(d, perm=[1, 0, 2])
            d = tf.reshape(d, [tf.shape(d)[0], -1])
            s = tf.convert_to_tensor([tf.cast(max_size - (end_batch - start_batch), dtype="int32"), tf.shape(d)[1]])
            d = tf.concat([d, tf.zeros(s, dtype="float32")], axis=0)
            return d

        data = tf.map_fn(extract_batch, [start_times, end_times], dtype=data.dtype)

        self.output.size_placeholder = {}
        self.output.size_placeholder[0] = new_sizes
        self.output.batch_dim_axis = 0
        self.output.time_dim_axis = 1
        self.output.placeholder = data

    @classmethod
    def get_out_data_from_opts(cls, name, sources, **kwargs):
        out = get_concat_sources_data_template(sources, name="%s_output" % name).copy_as_batch_major()
        out.size_placeholder[0] = None
        return out


class FillUnusedMemoryLayer(_ConcatInputLayer):
    """
    Fills all unused entries in the time/batch/feature tensor with a constant
    """

    layer_class = "fill_unused"

    def __init__(self, fill_value=0.0, **kwargs):
        super(FillUnusedMemoryLayer, self).__init__(**kwargs)

        mask = self.input_data.get_sequence_mask()
        mask = tf.expand_dims(mask, dim=-1)
        mask = tf.tile(mask, [1, 1, tf.shape(self.input_data.placeholder)[2]])

        x = self.input_data.placeholder
        x = tf.where(mask, x, tf.fill(tf.shape(x), float(fill_value)))
        self.output.placeholder = x
        self.output.size_placeholder = self.input_data.size_placeholder.copy()

    @classmethod
    def get_out_data_from_opts(cls, name, sources=(), **kwargs):
        return get_concat_sources_data_template(sources, name="%s_output" % name)


class SwapTimeFeatureLayer(_ConcatInputLayer):
    layer_class = "swap_time_feature"

    def __init__(self, **kwargs):
        super(SwapTimeFeatureLayer, self).__init__(**kwargs)
        assert self.input_data.batch_ndim == 3
        assert not self.input_data.sparse
        perm = [self.input_data.batch_dim_axis, self.input_data.feature_dim_axis, self.input_data.time_dim_axis]
        self.output.placeholder = tf.transpose(self.output.placeholder, perm=perm)
        shape = tf.shape(self.output.placeholder)
        self.output.size_placeholder[0] = tf.fill((shape[0],), value=shape[1])
        self.output.dim = self.output.placeholder.get_shape()[-1].value
        self.output.shape = (self.output.placeholder.get_shape()[1].value, self.output.dim)

    @classmethod
    def get_out_data_from_opts(cls, name, sources=(), **kwargs):
        out = get_concat_sources_data_template(sources, name="%s_output" % name)
        out.batch_dim_axis = 0
        out.time_dim_axis = 1
        out.dim = None
        out.size_placeholder = {}
        return out


class FlattenTimeLayer(_ConcatInputLayer):
    layer_class = "flatten_time"

    def __init__(self, **kwargs):
        super(FlattenTimeLayer, self).__init__(**kwargs)
        out = self.input_data.get_placeholder_as_batch_major()
        out = tf.reshape(out, [tf.shape(out)[0], 1, -1])  # (B, 1, T*D)
        self.output.placeholder = out
        self.output.size_placeholder[0] = tf.ones_like(self.input_data.get_sequence_lengths())

    @classmethod
    def get_out_data_from_opts(cls, name, sources, **kwargs):
        out = get_concat_sources_data_template(sources, name="%s_output" % name).copy_as_batch_major()
        out.dim = None
        out.shape = (1, None)
        out.size_placeholder = {}
        return out


class ApplyLengthDistributionLayer(LayerBase):
    layer_class = "apply_length_distribution"

    def __init__(self, length_model_scale=1.0, **kwargs):
        super(ApplyLengthDistributionLayer, self).__init__(**kwargs)
        self.output = self.sources[0].output.copy()
        len_dist_layer = self.sources[1]
        perm = []
        if self.output.is_batch_major:
            perm.append(len_dist_layer.output.batch_dim_axis)
            perm.append(len_dist_layer.output.feature_dim_axis)
        else:
            perm.append(len_dist_layer.output.feature_dim_axis)
            perm.append(len_dist_layer.output.batch_dim_axis)
        perm.append(len_dist_layer.output.time_dim_axis)
        len_mod = tf.transpose(len_dist_layer.output.placeholder, perm=perm)
        if length_model_scale != 1.0:
            len_mod = tf.pow(len_mod, length_model_scale)
        self.output.placeholder *= len_mod

    @classmethod
    def get_out_data_from_opts(cls, name, sources, **kwargs):
        return sources[0].output.copy()


class NormalizeLengthScoresLayer(LayerBase):
    layer_class = "normalize_length_scores"

    def __init__(self, **kwargs):
        super(NormalizeLengthScoresLayer, self).__init__(**kwargs)
        time_axis = 0 if self.sources[0].output.is_time_major else 1
        batch_axis = 1 if self.sources[0].output.is_time_major else 0

        self.output = self.sources[0].output.copy()
        p = self.sources[0].output.placeholder
        win_size = tf.cast(tf.shape(p)[time_axis], dtype=tf.float32)
        s = tf_compat.v1.log(p) * tf.expand_dims(
            tf.expand_dims(tf.range(win_size, dtype=tf.float32) + 1.0, -1), batch_axis
        )
        self.output.placeholder = tf.exp(s)

    @classmethod
    def get_out_data_from_opts(cls, name, sources, **kwargs):
        return sources[0].output.copy()



#!/usr/bin/env python3

"""
RETURNN as a tool. Main entry point. Just calls :func:`returnn.__main__.main`.
"""

from returnn.__main__ import main


if __name__ == "__main__":
    main()

else:
    # This is likely an old script, directly importing this file.
    # You should not do that. Import returnn.__main__ instead.
    # Anyway, here some hacky workaround to keep this working.
    import sys

    sys.modules[__name__] = sys.modules["returnn.__main__"]



"""
test threading-specific functionality
"""

from __future__ import annotations

import _setup_test_env  # noqa
from subprocess import Popen, PIPE, STDOUT, CalledProcessError
import os
import sys
import unittest
import tempfile
import textwrap
from returnn.util import better_exchook


__my_dir__ = os.path.dirname(os.path.abspath(__file__))
__base_dir__ = os.path.dirname(__my_dir__)
__main_entry__ = __base_dir__ + "/rnn.py"
py = sys.executable


def run(args, input=None):
    """run subproc"""
    args = list(args)
    print("run:", args)
    # RETURNN by default outputs on stderr, so just merge both together
    p = Popen(args, stdout=PIPE, stderr=STDOUT, stdin=PIPE)
    out, _ = p.communicate(input=input)
    print("Return code is %i" % p.returncode)
    print("std out/err:\n---\n%s\n---\n" % out.decode("utf8"))
    if p.returncode != 0:
        raise CalledProcessError(cmd=args, returncode=p.returncode, output=out)
    return out.decode("utf8")


def test_thread_exc_hook():
    config = textwrap.dedent(
        """
        #!rnn.py

        backend = "torch"  # just require any backend
        log_verbosity = 5
        task = "nop"

        import threading

        t = threading.Thread(target=lambda: 1/0)
        t.start()
        t.join()
        """
    )
    with tempfile.NamedTemporaryFile(mode="w") as f:
        f.write(config)
        f.flush()

        try:
            output = run([py, __main_entry__, f.name])
        except CalledProcessError as exc:
            out = exc.output.decode("utf8")
            assert "ZeroDivisionError" in out
            assert "KeyboardInterrupt" in out  # this is the result from the main thread being killed
        else:
            assert False, f"Expected RETURNN to crash, but got {output}."


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute
    #  better_exchook.dump_all_thread_tracebacks()



"""
What we want: subprocess.Popen to always work.
Problem: It uses fork+exec internally in subprocess_fork_exec, via _posixsubprocess.fork_exec.
That is a problem because fork can trigger any atfork handlers registered via pthread_atfork,
and those can crash/deadlock in some cases.

https://github.com/tensorflow/tensorflow/issues/13802
https://github.com/xianyi/OpenBLAS/issues/240
https://trac.sagemath.org/ticket/22021
https://bugs.python.org/issue31814

"""

from __future__ import annotations

import _setup_test_env  # noqa
import os
import sys
from pprint import pprint


my_dir = os.path.dirname(os.path.abspath(__file__))

# language=C++
c_code_at_fork_demo = """
#include <stdio.h>
#include <pthread.h>
// int pthread_atfork(void (*prepare)(void), void (*parent)(void), void (*child)(void));

static long magic_number = 0;

void set_magic_number(long i) {
    magic_number = i;
}

void hello_from_child() {
    printf("Hello from child atfork, magic number %li.\\n", magic_number);
    fflush(stdout);
}

void hello_from_fork_prepare() {
    printf("Hello from atfork prepare, magic number %li.\\n", magic_number);
    fflush(stdout);
}

void register_hello_from_child() {
    pthread_atfork(0, 0, &hello_from_child);
}

void register_hello_from_fork_prepare() {
    pthread_atfork(&hello_from_fork_prepare, 0, 0);
}
"""


class CLibAtForkDemo:
    def __init__(self):
        self._load_lib()

    def _load_lib(self):
        from returnn.util.basic import NativeCodeCompiler

        native = NativeCodeCompiler(
            base_name="test_fork_exec", code_version=1, code=c_code_at_fork_demo, is_cpp=False, ld_flags=["-lpthread"]
        )
        self._lib = native.load_lib_ctypes()
        print("loaded lib:", native.get_lib_filename())
        import ctypes

        self._lib.register_hello_from_child.restype = None  # void
        self._lib.register_hello_from_child.argtypes = ()
        self._lib.register_hello_from_fork_prepare.restype = None  # void
        self._lib.register_hello_from_fork_prepare.argtypes = ()
        self._lib.set_magic_number.restype = None  # void
        self._lib.set_magic_number.argtypes = (ctypes.c_long,)

    def set_magic_number(self, i):
        self._lib.set_magic_number(i)

    def register_hello_from_child(self):
        self._lib.register_hello_from_child()

    def register_hello_from_fork_prepare(self):
        self._lib.register_hello_from_fork_prepare()


clib_at_fork_demo = CLibAtForkDemo()


def demo_hello_from_fork():
    print("Hello.")
    sys.stdout.flush()
    clib_at_fork_demo.set_magic_number(3)
    clib_at_fork_demo.register_hello_from_child()
    clib_at_fork_demo.register_hello_from_fork_prepare()
    pid = os.fork()
    if pid == 0:
        print("Hello from child after fork.")
        sys.exit()
    print("Hello from parent after fork.")
    os.waitpid(pid, 0)
    print("Bye.")


def demo_start_subprocess():
    print("Hello.")
    sys.stdout.flush()
    clib_at_fork_demo.set_magic_number(5)
    clib_at_fork_demo.register_hello_from_child()
    clib_at_fork_demo.register_hello_from_fork_prepare()
    from subprocess import check_call

    # Right now (2017-10-19), CPython will use subprocess_fork_exec which
    # uses fork+exec, so this will call the atfork handler.
    check_call("echo Hello from subprocess.", shell=True)
    print("Bye.")


def run_demo_check_output(name):
    """
    :param str name: e.g. "demo_hello_from_fork"
    :return: lines of stdout of the demo
    :rtype: list[str]
    """
    from subprocess import check_output

    output = check_output([sys.executable, __file__, name])
    return output.decode("utf8").splitlines()


def filter_demo_output(ls):
    """
    :param list[str] ls:
    :rtype: list[str]
    """
    ls = [l for l in ls if not l.startswith("Executing: ")]
    ls = [l for l in ls if not l.startswith("Compiler call: ")]
    ls = [l for l in ls if not l.startswith("loaded lib: ")]
    ls = [l for l in ls if not l.startswith("dlopen: ")]
    ls = [
        l
        for l in ls
        if "installLibSigSegfault" not in l and "libSegFault" not in l and "native_signal_handler" not in l
    ]
    ls = [l for l in ls if "faulthandler" not in l]
    found_hello = False
    for i, l in enumerate(ls):
        # Those can be very early, before the hello.
        if l in ["Ignoring pthread_atfork call!", "Ignoring __register_atfork call!"]:
            continue
        assert l == "Hello."
        ls.pop(i)
        found_hello = True
        break
    assert found_hello, "no Hello: %r" % (ls,)
    assert ls[-1] == "Bye."
    ls = ls[:-1]
    return ls


def test_demo_hello_from_fork():
    ls = run_demo_check_output("demo_hello_from_fork")
    pprint(ls)
    ls = filter_demo_output(ls)
    pprint(ls)
    assert set(ls) == {
        "Hello from child after fork.",
        "Hello from child atfork, magic number 3.",
        "Hello from atfork prepare, magic number 3.",
        "Hello from parent after fork.",
    }


def test_demo_start_subprocess():
    ls = run_demo_check_output("demo_start_subprocess")
    pprint(ls)
    ls = filter_demo_output(ls)
    pprint(ls)
    assert "Hello from subprocess." in ls
    # Maybe this is fixed/changed in some later CPython version.
    import platform

    print("Python impl:", platform.python_implementation())
    print("Python version:", sys.version_info[:3])
    if platform.python_implementation() == "CPython":
        if sys.version_info[0] == 2:
            old = sys.version_info[:3] <= (2, 7, 12)
        else:
            old = sys.version_info[:3] <= (3, 6, 1)
    else:
        old = False
    if old:
        print("atfork handler should have been called.")
        assert "Hello from child atfork, magic number 5." in ls
        assert "Hello from atfork prepare, magic number 5." in ls
    else:
        print("Not checking for atfork handler output.")


def patched_check_demo_start_subprocess():
    """
    Just like test_demo_start_subprocess(), but here we assert that no atfork handlers are executed.
    """
    assert os.environ.get("__RETURNN_ATFORK_PATCHED") == "1"
    ls = run_demo_check_output("demo_start_subprocess")
    pprint(ls)
    ls = filter_demo_output(ls)
    pprint(ls)
    assert "Hello from subprocess." in ls
    ls = [l for l in ls if l not in ["Ignoring pthread_atfork call!", "Ignoring __register_atfork call!"]]
    pprint(ls)
    assert ls == ["Hello from subprocess."]


def test_demo_start_subprocess_patched():
    from returnn.util.basic import get_patch_atfork_lib
    from subprocess import check_call

    env = os.environ.copy()
    env["LD_PRELOAD"] = get_patch_atfork_lib()
    print("LD_PRELOAD:", get_patch_atfork_lib())
    check_call([sys.executable, __file__, "patched_check_demo_start_subprocess"], env=env)


if __name__ == "__main__":
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                v()
                print("-" * 40)
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



from __future__ import annotations

import _setup_test_env  # noqa
import os
import tempfile
import atexit
import shutil
import pickle
import numpy

from returnn.datasets import init_dataset
from returnn.datasets.huggingface import HuggingFaceDataset
from test_Dataset import dummy_iter_dataset


def _setup_hf_env():
    if "HF_HOME" not in os.environ:
        os.environ["HF_HOME"] = _get_tmp_dir()


def _get_tmp_dir() -> str:
    fn = tempfile.mkdtemp()
    atexit.register(shutil.rmtree, fn)
    return fn


_setup_hf_env()


def test_HuggingFaceDataset_audio():
    ds = HuggingFaceDataset(
        {"path": "datasets-examples/doc-audio-6", "split": "train"},
        cast_columns={"audio": {"_type": "Audio", "sample_rate": 16_000}},
        data_format={"audio": {"dtype": "float32", "shape": [None]}},
        seq_tag_column=None,
    )
    ds.initialize()
    res = dummy_iter_dataset(ds)
    print(res[0].features["audio"])


def test_HuggingFaceDataset_text1():
    ds = HuggingFaceDataset(
        {"path": "openai/gdpval", "split": "train"},
        seq_tag_column="task_id",
        data_format={
            "prompt": {"dtype": "string", "shape": ()},
            "sector": {"dtype": "string", "shape": ()},
            "occupation": {"dtype": "string", "shape": ()},
        },
    )
    ds.initialize()
    res = dummy_iter_dataset(ds)
    print(repr(res[0].seq_tag))
    assert type(res[0].seq_tag) is str


def test_HuggingFaceDataset_text2():
    ds = HuggingFaceDataset(
        {"path": "lavita/medical-qa-shared-task-v1-toy", "split": "train"},
        seq_tag_column="id",
        data_format={
            "id": {"dtype": "int64", "shape": ()},
            "startphrase": {"dtype": "string", "shape": ()},
            "label": {"dtype": "int64", "shape": ()},
        },
    )
    ds.initialize()
    assert dummy_iter_dataset(ds)


def test_HuggingFaceDataset_rename_tokens():
    ds = HuggingFaceDataset(
        {"path": "lavita/medical-qa-shared-task-v1-toy", "split": "train"},
        seq_tag_column="id",
        rename_columns={"startphrase": "text"},
        data_format={
            "id": {"dtype": "int64", "shape": ()},
            "text": {"dtype": "string", "shape": ()},
            "label": {"dtype": "int64", "shape": ()},
        },
    )
    ds.initialize()
    assert dummy_iter_dataset(ds)


def test_HuggingFaceDataset_text_tokenize():
    ds = HuggingFaceDataset(
        {"path": "lavita/medical-qa-shared-task-v1-toy", "split": "train"},
        seq_tag_column="id",
        data_format={
            "id": {"dtype": "int64", "shape": ()},
            "startphrase": {"dtype": "int32", "vocab": {"class": "Utf8ByteTargets"}},
            "label": {"dtype": "int64", "shape": ()},
        },
    )
    ds.initialize()
    res = dummy_iter_dataset(ds)
    txt = res[0].features["startphrase"]
    print("startphrase:", txt)
    assert isinstance(txt, numpy.ndarray) and txt.dtype == numpy.int32
    txt_ = ds.data_format["startphrase"].vocab.get_seq_labels(txt)
    print("startphrase labels:", txt_)


def test_HuggingFaceDataset_pickle():
    ds = HuggingFaceDataset(
        {"path": "lavita/medical-qa-shared-task-v1-toy", "split": "train"},
        seq_tag_column="id",
        data_format={
            "id": {"dtype": "int64", "shape": ()},
            "startphrase": {"dtype": "string", "shape": ()},
            "label": {"dtype": "int64", "shape": ()},
        },
    )
    ds.initialize()
    s = pickle.dumps(ds)
    ds = pickle.loads(s)
    assert isinstance(ds, HuggingFaceDataset)
    assert dummy_iter_dataset(ds)


def test_HuggingFaceDataset_load_from_disk():
    from datasets import load_dataset

    datadir_path = _get_tmp_dir() + "/hf-dataset-save-to-disk"
    hf_ds = load_dataset("lavita/medical-qa-shared-task-v1-toy", split="train")
    hf_ds.save_to_disk(datadir_path)

    ds = HuggingFaceDataset(
        datadir_path,
        seq_tag_column="id",
        data_format={
            "id": {"dtype": "int64", "shape": ()},
            "startphrase": {"dtype": "string", "shape": ()},
            "label": {"dtype": "int64", "shape": ()},
        },
    )
    ds.initialize()
    assert dummy_iter_dataset(ds)


def test_HuggingFaceDataset_single_arrows():
    import datasets

    datadir_path = _get_tmp_dir() + "/hf-dataset-save-to-disk"
    hf_ds = datasets.Dataset.from_list([{"data": i} for i in range(100_000)])
    hf_ds.save_to_disk(datadir_path, num_shards=100)

    content = os.listdir(datadir_path)
    print("Saved dir content:", content)
    assert "state.json" in content
    assert "dataset_info.json" in content
    assert all(f"data-{i:05}-of-00100.arrow" in content for i in range(100))

    ds = HuggingFaceDataset(
        [f"{datadir_path}/data-{i:05}-of-00100.arrow" for i in range(0, 100, 5)],
        seq_tag_column=None,
        data_format={"data": {"dtype": "int64", "shape": ()}},
    )
    ds.initialize()
    assert dummy_iter_dataset(ds)


def test_HuggingFaceDataset_file_cache_with_sharded():
    import datasets

    datadir_path = _get_tmp_dir() + "/hf-dataset-save-to-disk"
    hf_ds = datasets.Dataset.from_list([{"data": i} for i in range(100_000)])
    hf_ds.save_to_disk(datadir_path, num_shards=100)

    ds = HuggingFaceDataset(
        datadir_path,
        use_file_cache=True,
        seq_tag_column=None,
        data_format={"data": {"dtype": "int64", "shape": ()}},
    )
    ds.initialize()
    assert dummy_iter_dataset(ds)


def test_HuggingFaceDataset_in_multi_proc():
    ds_dict = {
        "class": "HuggingFaceDataset",
        "dataset_opts": {"path": "lavita/medical-qa-shared-task-v1-toy", "split": "train"},
        "seq_tag_column": "id",
        "data_format": {
            "id": {"dtype": "int64", "shape": ()},
            "startphrase": {"dtype": "string", "shape": ()},
            "label": {"dtype": "int64", "shape": ()},
        },
    }
    ds_dict = {
        "class": "MultiProcDataset",
        "num_workers": 2,
        "buffer_size": 5,
        "dataset": ds_dict,
    }
    ds = init_dataset(ds_dict)
    ds.initialize()
    assert dummy_iter_dataset(ds)
